from .action_core import *
from .action_variants import *
from .combinations import *
from ..expressions.operations import *
from ..expressions.variables import *
from ..expressions.relations import *
from MF_Tools.dual_compatibility import PI, FadeIn


class swap_children_(Action):
    def __init__(self, preaddress='', mode="arc", arc_size=0.75*PI, **kwargs):
        self.mode = mode
        self.arc_size = arc_size
        super().__init__(preaddress=preaddress,**kwargs)
    
    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        assert len(input_expression.children) == 2, f"Cannot swap children of {input_expression}, must have two children."
        return type(input_expression)(input_expression.children[1], input_expression.children[0])

    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        if self.mode == "arc":
            return [
                ["0", "1", {"path_arc": self.arc_size}],
                ["1", "0", {"path_arc": self.arc_size}]
            ]
        elif self.mode == "straight":
            return [
                ["0", "1"],
                ["1", "0"]
            ]
        else:
            raise ValueError(f"Invalid mode: {self.mode}. Must be 'arc' or 'straight'.")


class apply_operation_(Action):
    def __init__(self, OpClass, other, preaddress='', side="right", introducer=Write, **kwargs):
        self.OpClass = OpClass
        self.other = Smarten(other)
        self.side = side
        self.introducer = introducer
        super().__init__(preaddress=preaddress,**kwargs)
    
    @preaddressfunc
    def get_output_expression(self, input_expression):
        if self.side == "right":
            output_expression = self.OpClass(input_expression, self.other) # Putting a .copy() on the input expression fixes the problem. Try to understand this and then probably fold this into the decorator
        elif self.side == "left":
            output_expression = self.OpClass(self.other, input_expression)
        else:
            raise ValueError(f"Invalid side: {self.side}. Must be left or right.")
        return output_expression

    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression):
        if self.side == "right":
            return [
                ["", "0"],
                [self.introducer, "+", {"delay":0.5}],
                [self.introducer, "1", {"delay":0.6}]
            ]
        elif self.side == "left":
            return [
                ["", "1"],
                [self.introducer, "0", {"delay":0.5}],
                [self.introducer, "+", {"delay":0.6}]
            ]
        else:
            raise ValueError(f"Invalid side: {self.side}. Must be left or right.")
    
    def __repr__(self):
        return type(self).__name__ + "(" + str(self.other) + ',' + self.preaddress + ")"

class add_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Add, other, *args, **kwargs)

class sub_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Sub, other, *args, **kwargs)

class mul_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Mul, other, *args, **kwargs)

class div_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Div, other, *args, **kwargs)

class pow_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Pow, other, *args, **kwargs)

class equals_(apply_operation_):
    def __init__(self, other, *args, **kwargs):
        super().__init__(Equation, other, *args, **kwargs)


class substitute_(Action):
    def __init__(self, sub_dict, preaddress='', mode="transform", arc_size=PI, fade_shift=DOWN*0.2, lag=0, **kwargs):
        self.sub_dict = sub_dict
        self.preaddress = preaddress
        self.mode = mode
        self.arc_size = arc_size
        self.fade_shift = fade_shift
        self.lag = lag #usually looks like shit but can be cool sometimes
        super().__init__(preaddress=preaddress,**kwargs)
    
    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        return input_expression.substitute(self.sub_dict)

    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        target_addresses = []
        for var in self.sub_dict:
            target_addresses += input_expression.get_subex(self.preaddress).get_addresses_of_subex(var)
        addressmap = []
        if self.mode == "transform":
            for i,ad in enumerate(target_addresses):
                addressmap.append([ad, ad, {"delay": self.lag*i}])
            return addressmap
        elif self.mode == "swirl":
            for i,ad in enumerate(target_addresses):
                addressmap.append([ad, ad, {"path_arc": self.arc_size, "delay": self.lag*i}])
            return addressmap
        elif self.mode == "fade":
            for i,ad in enumerate(target_addresses):
                addressmap.append([ad, FadeOut, {"shift": self.fade_shift, "delay": self.lag*i}])
                addressmap.append([FadeIn, ad, {"shift": self.fade_shift, "delay": self.lag*i}])
            return addressmap
        
    def __repr__(self):
        return type(self).__name__ + "(" + str(self.sub_dict) + (',' + self.preaddress if self.preaddress else '') + ")"


class substitute_into_(Action):
    def __init__(self, outer_expression, substitution_variable=Variable('x'), **kwargs):
        self.outer_expression = outer_expression
        self.substitution_variable = substitution_variable
        super().__init__(**kwargs)
    
    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        return self.outer_expression.substitute({self.substitution_variable: input_expression})
    
    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        addressmap = []
        sub_into_addresses = self.outer_expression.get_addresses_of_subex(self.substitution_variable)
        for ad in sub_into_addresses:
            if not input_expression.parentheses:
                ad = ad + "_"
            addressmap.append(['', ad])
        return addressmap
    
    def get_animation(self, *args, **kwargs):
        return super().get_animation(*args, auto_fade=True, auto_resolve_delay=0.75, **kwargs)


class evaluate_(Action):
    def __init__(self, preaddress='', mode="random leaf", **kwargs):
        super().__init__(preaddress=preaddress,**kwargs)
        # if mode == "random leaf":
        #     leaf_addresses = input_expression.get_all_leaf_addresses()
        #     leaf_address = np.random.choice(leaves)
        #     self.preaddress = leaf_address

    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        if isinstance(input_expression, Relation):
            raise ValueError("Not implemented yet")
        return input_expression.evaluate()
    
    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        return [
            ["", ""] #extension by preaddress is done by decorator!
        ]
    

class distribute_(Action):
       # Not done yet, multilayer does not work... this is so necessary but rather nontrivial... hm...
    def __init__(self, preaddress='', mode="auto", multilayer=False, **kwargs):
        self.mode = mode #"auto", "left", "right"
        super().__init__(preaddress=preaddress,**kwargs)
    
    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        if self.mode == "auto":
            self.determine_direction(input_expression)
        if self.mode == "left":
            new_children = [
                type(input_expression)(input_expression.children[0], child)
                for child in input_expression.children[-1].children
                ]
            return type(input_expression.children[-1])(*new_children)
        elif self.mode == "right":
            new_children = [
                type(input_expression)(child, input_expression.children[-1])
                for child in input_expression.children[0].children
                ]
            return type(input_expression.children[0])(*new_children)

    def determine_direction(self, input_expression=None):
        if self.mode == "auto":
            if isinstance(input_expression, Mul):
                left_distributable = isinstance(input_expression.children[-1], (Add, Sub))
                right_distributable = isinstance(input_expression.children[0], (Add, Sub))
                if left_distributable and right_distributable:
                    raise ValueError("Cannot auto-distribute if both sides are distributable, please set mode manually.")
                elif left_distributable:
                    self.mode = "left"
                elif right_distributable:
                    self.mode = "right"
                else:
                    raise ValueError("Cannot distribute, neither side is distributable.")
            elif isinstance(input_expression, Div):
                right_distributable = isinstance(input_expression.children[0], (Add, Sub))
                if right_distributable:
                    self.mode = "right"
                else:
                    raise ValueError("Cannot distribute, right side is not distributable.")
            elif isinstance(input_expression, Pow):
                right_distributable = isinstance(input_expression.children[0], (Mul, Div))
                if right_distributable:
                    self.mode = "right"
                else:
                    raise ValueError("Cannot distribute, right side is not distributable.")
            else:
                raise ValueError("Cannot auto-distribute, must be a multiplication or division.")

    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        return [
            ["", ""] #standin idk what the fuck im doing here
        ]






# from .action_variants import AlgebraicAction
# a = Variable('a')
# b = Variable('b')
# c = Variable('c')

# square_binomial_ = AlgebraicAction((a+b)**2, a**2 + 2*a*b + b**2)

# AlgebraMoves = [
#     AlgebraicAction(a+b & c, a & c-b, var_kwarg_dict={b:{"path_arc":PI}}, extra_addressmaps=[['0+', '1-', {'path_arc': PI}]]),
#     AlgebraicAction(a-b & c, a & c+b),
#     AlgebraicAction(a*b & c, b & c/a, var_kwarg_dict={a:{"path_arc":PI*2/3}}, extra_addressmaps=[[Write, '1/', {'delay':0.25}]]),
#     AlgebraicAction(a/b & c, a & c*b, var_kwarg_dict={b:{"path_arc":PI*2/3}}, extra_addressmaps=[['0/', FadeOut, {'delay':0}], [FadeIn, '11()', {'delay':1, 'run_time':0.5}]]),
# ]



__version__ = '0.5.1'
__author__ = 'John Connell - The Mathematic Fanatic'
__description__ = 'Manim plugin which aims to make it much easier to meaningfully transform algebra expressions.'

from .expressions import *
from .actions import *
from .timelines import *



from ...expressions.functions import *
from ...expressions.variables import Variable
import numpy as np


theta = Variable("\\theta")

sin = Function("\\sin", 3, rule=np.sin, parentheses_mode="weak")
cos = Function("\\cos", 3, rule=np.cos, parentheses_mode="weak")
tan = Function("\\tan", 3, rule=np.tan, parentheses_mode="weak")
cot = Function("\\cot", 3, rule=lambda x: 1/np.tan(x), parentheses_mode="weak")
sec = Function("\\sec", 3, rule=lambda x: 1/np.cos(x), parentheses_mode="weak")
csc = Function("\\csc", 3, rule=lambda x: 1/np.sin(x), parentheses_mode="weak")




from .expression_core import *
from .operations import *
from .numbers import *
from .variables import *
from .functions import *
from .sequences import *
from .relations import *
from .expression_common import *


# actions.py
from MF_Tools.dual_compatibility import Write, FadeIn, FadeOut
from ..expressions.expression_core import *
from ..utils import *
from .animations import TransformByAddressMap


class Action:
	"""
		Transforms Expressions into other Expressions,
		both as static objects and also with an animation.

		An action is defined by two main things:
		the get_output_expression method, which controls how it acts on static expressions,
		and the get_addressmap method, which controls how it acts as an animation.
		Both set attributes of the corresponding name and return them.

		It may also have a preaddress parameter/attribute which will determine the subexpression
		address at which the action is applied, and a few other attributes which may adjust some
		specifics.

		self.input_expression is set to None during __init__. It is critical that actions
		can exist prior to being given expressions, so that they can be combined together.
		When an input expression is received, this attribute is set, and the method
		.get_output_expression is called, setting self.output_expression.
		This is all that is required for static actions, no animations.

		Now, to create the animation between these expressions:

		get_addressmap is also unique to each action, and returns something like
		[
			["00", "01"],
			["01", "00", {"path_arc":PI/2}],
			[FadeIn, "1"],
			["1", FadeOut]
		]
		which contains all the expression-agnostic information about the animation.
		Often this will simply define and return this list with no computation.

		get_glyphmap combines the input_expression, output_expression, and addressmap
		to create a list like
		[
			([0,1,2], [5,6]),
			([3,4,5], [1,2,3], {"path_arc":PI/2}),
			(FadeIn, [8,9]),
			([6], FadeOut)
		]
		which tells which glyphs of the mobjects to send to which others, and how.

		get_animation then simply parses this glyphmap list to create a list of
		animations, probably to be passed to AnimationGroup, like
		[
			ReplacementTransform(A[0][0,1,2], B[0][5,6]),
			ReplacementTransform(A[0][3,4,5], B[0][1,2,3], path_arc=PI/2),
			FadeIn(B[0][8,9]),
			FadeOut(A[0][6]),
			...
		]
		or something like that, the syntax is partially made up. The ... is
		ReplacementTransforms of all the individual glyphs not mentioned in the glyphmap,
		whose lengths have to exactly match.

		Broadly speaking, that's that!
	"""
	def __init__(self,
		introducer=Write,
		remover=FadeOut,
		preaddress=''
	):
		self.introducer = introducer
		self.remover = remover
		self.preaddress = preaddress

	def get_output_expression(self, input_expression):
  		# define in subclasses and decorate with @preaddressfunc
		raise NotImplementedError

	def get_addressmap(self, input_expression, **kwargs):
		# define in subclasses and decorate with @preaddressmap
		raise NotImplementedError

	def get_animation(self, **kwargs):
		def animation(input_exp, output_exp=None):
			if output_exp is None:
				output_exp = self.get_output_expression(input_exp)
			return TransformByAddressMap(
			input_exp,
			output_exp,
			*self.get_addressmap(input_exp),
			default_introducer=self.introducer,
			default_remover=self.remover,
			**kwargs
			)
		return animation
	
	def __call__(self, expr1, expr2=None, **kwargs):
		return self.get_animation(**kwargs)(expr1, expr2)

	def __or__(self, other):
		from .combinations import ParallelAction
		if isinstance(other, ParallelAction):
			return ParallelAction(self, *other.actions)
		elif isinstance(other, Action):
			return ParallelAction(self, other)
		else:
			raise ValueError("Can only use | with other ParallelAction or Action")
	
	def __rshift__(self, other):
		other = Smarten(other)
		from ..expressions.expression_core import Expression
		from ..timelines.timeline_core import Timeline
		if isinstance(other, Expression):
			timeline = Timeline()
			timeline.add_action_to_end(self).add_expression_to_end(other)
			return timeline
		elif isinstance(other, Action):
			timeline = Timeline()
			timeline.add_action_to_end(self).add_action_to_end(other)
			return timeline
		else:
			return NotImplemented

	def __rrshift__(self, other):
		return Smarten(other).__rshift__(self)
	
	def __repr__(self):
		return type(self).__name__ + "(" + self.preaddress + ")"
	
	def copy(self):
		return deepcopy(self)
	
	def both(self, number_of_sides=2):
		# Intended to turn an action on an expression into an action done to both sides of an equation.
		# Can be passed a number to apply to more than 2 sides for, say, a triple equation or inequality.
		return self.pread(*[str(i) for i in range(number_of_sides)])
	
	def pread(self, *addresses):
		if len(addresses) == 0:
			return self
		elif len(addresses) == 1:
			self.preaddress = addresses[0] + self.preaddress
			return self
		else:
			actions = []
			for ad in addresses:
				action = self.copy().pread(ad)
				actions.append(action)
			from .combinations import ParallelAction
			return ParallelAction(*actions)

	def __leq__(self, expr):
		assert isinstance(expr, Expression), "Can only apply expression >= action"
		return self.get_output_expression(expr)


def preaddressfunc(func):
	def wrapper(action, expr, *args, **kwargs):
		expr = expr.copy()
		preaddress = kwargs.get('preaddress', '') or action.preaddress
		if len(preaddress)==0:
			output_expression = func(action, expr)
		else:
			active_part = expr.get_subex(preaddress)
			result = func(action, active_part)
			output_expression = expr.substitute_at_address(result, preaddress)
		output_expression.reset_parentheses()
		return output_expression
	return wrapper

def preaddressmap(getmap):
	def wrapper(action, expr, *args, **kwargs):
		expr = expr.copy()
		preaddress = kwargs.get('preaddress', '') or action.preaddress
		addressmap = getmap(action, expr, *args, **kwargs)
		if preaddress:
			for entry in addressmap:
				for i, ad in enumerate(entry):
					if isinstance(ad, str):
						entry[i] = preaddress + ad
		return addressmap
	return wrapper

def autoparenmap(getmap, mode='stupid'):
	if mode=='stupid':
		def wrapper(action, expr, *args, **kwargs):
			in_expr, out_expr = expr, action.get_output_expression(expr)
			addressmap = list(getmap(action, expr, *args, **kwargs))
			for in_add in in_expr.get_all_addresses():
				if in_expr.get_subex(in_add).parentheses:
					addressmap.append([in_add+'()', FadeOut, {'run_time':0.5}])
				for entry in addressmap:
					if entry[0] == in_add:
						entry[0] = entry[0] + '_'
			for out_add in out_expr.get_all_addresses():
				if out_expr.get_subex(out_add).parentheses:
					addressmap.append([FadeIn, out_add+'()', {'run_time':0.5, 'delay':0.5}])
				for entry in addressmap:
					if entry[1] == out_add:
						entry[1] = entry[1] + '_'
			return addressmap

	if mode=='smart':
		def wrapper(action, expr, *args, **kwargs):
			expr = expr.copy()
			addressmap = getmap(action, expr, *args, **kwargs)
			active_from, active_to = [], []
			for entry in addressmap:
				pass
	return wrapper


class IncompatibleExpression(Exception):
	pass





  




 


from ..expressions import *
from ..actions import *
from MF_Tools.dual_compatibility import TransformMatchingShapes, UP, smooth
# from MF_Tools.rescaling import *


class Timeline:
    """
    A class that represents a timeline of expressions and actions.
    """
    def __init__(
        self,
        auto_color = {},
        auto_scale = 1,
        auto_fit = [None, None, None],
        auto_propagate = True,
        show_past_steps = False,
        past_steps_opacity = 0.4,
        past_steps_direction = UP,
        past_steps_buff = 1,
        past_steps_shift_run_time = 1,
        past_steps_shift_rate_func = smooth
    ):
        self.steps = [] # Elements of this list are of the form [expression, action]
        self.current_exp_index = 0
        self.auto_color = auto_color
        self.auto_scale = auto_scale
        self.auto_fit = auto_fit
        self.auto_propagate = auto_propagate
        self.show_past_steps = show_past_steps
        if self.show_past_steps:
            self.past_steps_vgroup = VGroup()
            self.past_steps_opacity = past_steps_opacity
            self.past_steps_direction = past_steps_direction
            self.past_steps_buff = past_steps_buff
            self.past_steps_shift_run_time = past_steps_shift_run_time
            self.past_steps_shift_rate_func = past_steps_shift_rate_func

    def get_expression(self, index: int) -> Expression:
        try:
            return self.steps[index][0]
        except IndexError:
            return None
    
    def set_expression(self, index: int, expression: Expression):
        if self.auto_color:
            expression.set_color_by_subex(self.auto_color)
        if any(self.auto_fit):
            expression.mob.scale_to_fit(*self.auto_fit)
        elif self.auto_scale != 1:
            expression.mob.scale(self.auto_scale)
        if index == len(self.steps):
            self.add_expression_to_end(expression)
        self.steps[index][0] = expression
        if self.auto_propagate:
            self.propagate(start_at=index)

    def add_expression_to_start(self, expression: Expression):
        if len(self.steps) == 0 or self.steps[0][0] is not None:
            self.steps.insert(0, [None, None])
        self.set_expression(0, expression)
        return self

    def add_expression_to_end(self, expression: Expression):
        self.steps.append([None, None])
        self.set_expression(-1, expression)
        return self
    
    def get_action(self, index: int) -> Action:
        try:
            return self.steps[index][1]
        except IndexError:
            return None

    def set_action(self, index: int, action: Action):
        self.steps[index][1] = action
        if self.auto_propagate:
            self.propagate(start_at=index)
    
    def add_action_to_start(self, action: Action):
        self.set_action(0, action)
        return self
    
    def add_action_to_end(self, action: Action):
        if len(self.steps) == 0 or self.steps[-1][1] is not None:
            self.steps.append([None, None])
        self.set_action(-1,action)
        return self
    
    def propagate(self, start_at=0):
        for i in range(start_at, len(self.steps) - 1):
            exp, act = self.steps[i]
            next_exp = self.steps[i+1][0]
            if exp != None and act != None and next_exp == None:
                self.set_expression(i+1, act.get_output_expression(exp))
        exp, act = self.steps[-1]
        if exp != None and act != None:
            try:
                self.add_expression_to_end(act.get_output_expression(exp))
            except NotImplementedError:
                pass
    

    def get_animation(self, index, **kwargs):
        action = self.get_action(index)
        expA = self.get_expression(index)
        expB = self.get_expression(index+1)
        if action:
            Animation = action.get_animation()(expA, expB, **kwargs)
        else:
            Animation = TransformMatchingShapes(expA.mob, expB.mob, **kwargs)
        return Animation

    def play_animation(self, scene, index, **kwargs):
        expA = self.get_expression(index)
        expB = self.get_expression(index+1)
        if self.show_past_steps:
            self.shift_past_steps(scene, expA, expB)
        animation = self.get_animation(index, **kwargs)
        scene.play(animation)
        self.current_exp_index = index+1
    
    def play_next(self, scene):
        self.play_animation(scene, index=self.current_exp_index)

    def play_range(self, scene, start_index, end_index, wait_between=1, **kwargs):
        for i in range(start_index, end_index):
            self.play_animation(scene, index=i, **kwargs)
            scene.wait(wait_between)
    
    def play_all(self, scene, wait_between=1, reset_exp_index=False):
        if reset_exp_index:
            self.current_exp_index = 0
        if self.mob not in scene.mobjects:
            scene.play(Write(self.mob))
            scene.wait(wait_between)
        while self.current_exp_index < len(self.steps)-1:
            self.play_next(scene=scene)
            scene.wait(wait_between)
    
    def shift_past_steps(self, scene, expA, expB):
        mobA_radius = expA.mob.get_critical_point(self.past_steps_direction) - expA.mob.get_center()
        mobB_radius = expB.mob.get_center() - expB.mob.get_critical_point(-self.past_steps_direction)
        shift_distance = np.linalg.norm(mobA_radius) + np.linalg.norm(mobB_radius) + self.past_steps_buff
        self.past_steps_vgroup.add(
            self.mob.copy().set_opacity(0.25)
        )
        scene.add(self.past_steps_vgroup)
        scene.play(
            self.past_steps_vgroup.animate.shift(shift_distance * self.past_steps_direction),
            run_time = self.past_steps_shift_run_time,
            rate_func = self.past_steps_shift_rate_func
        )
    
    def __rshift__(self, other):
        if isinstance(other, Expression):
            self.add_expression_to_end(other)
            return self
        elif isinstance(other, Action):
            self.add_action_to_end(other)
            return self
        elif isinstance(other, Timeline):
            raise NotImplementedError("TODO")
        else:
            raise ValueError('Timeline can only be combined via >> with a Expression, Action, or another Timeline')
    
    def __rrshift__(self, other):
        if isinstance(other, Expression):
            return self.add_expression_to_start(other)
        elif isinstance(other, Action):
            return self.add_action_to_start(other)
        elif isinstance(other, Timeline):
            raise NotImplementedError("TODO")
        else:
            raise ValueError('Timeline can only be combined via >> with a Expression, Action, or another Timeline')
        
    def __repr__(self):
        return f"Timeline({self.steps})"

    def get_vgroup(self, **kwargs):
        return VGroup(*[self.steps[i][0].mob for i in range(len(self.steps))])
    
    @property
    def mob(self):
        return self.steps[self.current_exp_index][0].mob
    





from .expression_core import *


class Function(Expression):
	def __init__(self,
		symbol,
		symbol_glyph_length = None,
		python_rule = None,
		algebra_rule_variables = [],
		algebra_rule = None,
		parentheses_mode="always",
		spacing = "",
		**kwargs
	):
		self.symbol = symbol #string
		self.symbol_glyph_length = symbol_glyph_length #int
		self.python_rule = python_rule #callable
		self.algebra_rule_variables = algebra_rule_variables
		self.algebra_rule = algebra_rule
		self.parentheses_mode = parentheses_mode
		self.spacing = spacing
		# First child is argument(s) such as a Variable, Number, or Sequence.
		# Further children are parameters like subscripts, indices, or bounds.
		self.children = []
		super().__init__(**kwargs)
	
	@tex
	def __str__(self):
		symbol_string = self.get_symbol_string()
		argument_string = str(self.children[0]) if len(self.children) > 0 else ""
		return symbol_string + self.spacing + '{' + argument_string + '}'
	
	def get_symbol_string(self):
		# Overwrite for subclasses with indices, subscripts, etc
		return self.symbol

	def __call__(self, *inputs):
		new_func = self.copy()
		if len(inputs) == 1:
			new_func.children.append(Smarten(inputs[0]))
		elif len(inputs) > 1:
			from .sequences import Sequence
			new_func.children.append(Sequence(*list(map(Smarten, inputs))))
		new_func.auto_parentheses()
		new_func._mob = None
		return new_func
	
	def auto_parentheses(self):
		if len(self.children) == 0:
			return self
		from ..expressions.sequences import Sequence
		if self.parentheses_mode == 'always' or isinstance(self.children[0], Sequence):
			self.children[0].give_parentheses(True)
			return self
		from ..expressions.operations import Operation, Add, Sub
		from ..expressions.functions import Function
		if self.parentheses_mode == 'strong' and isinstance(self.children[0], (Operation, Function)):
			self.children[0].give_parentheses(True)
		if self.parentheses_mode == 'weak' and isinstance(self.children[0], (Add, Sub)):
			self.children[0].give_parentheses(True)
		if self.parentheses_mode == 'never':
			self.children[0].give_parentheses(False)
		return self
	
	def compute(self):
		from ..expressions.sequences import Sequence
		if len(self.children) == 0:
			raise ValueError(f"Function {self.symbol} has no arguments.")
		if isinstance(self.children[0], Sequence):
			args = self.children[0].children
		else:
			args = [self.children[0]]
		if self.python_rule is not None:
			return self.python_rule(*args)
		elif self.algebra_rule is not None:
			if self.algebra_rule_variables is not None:
				substituted_expression = self.algebra_rule.substitute({var:val for var, val in zip(self.algebra_rule_variables, args)})
			elif len(var_set := self.algebra_rule.get_all_variables()) == len(args) == 1:
				substituted_expression = self.algebra_rule.substitute({list(var_set):args[0]})
			else:
				raise ValueError(f"Algebra rule {self.algebra_rule} requires {len(self.algebra_rule_variables)} arguments, but {len(args)} were given.")
			return substituted_expression.compute()



class _OldFunction(Expression):
	def __init__(self, symbol, symbol_glyph_length, rule=None, algebra_rule=None, parentheses_mode="always", **kwargs):
		self.symbol = symbol #string
		self.symbol_glyph_length = symbol_glyph_length #int
		self.rule = rule #callable
		self.children = [Sequence()] # First child is always a sequence of arguments, further children are parameters like subscripts or indices
		self.algebra_rule = algebra_rule #SmE version of rule?
		self.parentheses_mode = parentheses_mode
		self.spacing = ""
		super().__init__(**kwargs)

	@tex
	def __str__(self):
		return self.symbol + self.spacing + (str(self.children[0]) if len(self.children) > 0 else "")

	def __call__(self, *inputs, **kwargs):
		assert len(self.children[0].children) == 0, f"Function {self.symbol} cannot be called because it already has children."
		new_func = self.copy()
		new_func.children[0].children = list(map(Smarten, inputs))
		# have to reinitialize Expression and MathTex after setting children for correct indexing and auto_paren.
		Expression.__init__(self, parentheses = self.parentheses)
		return new_func
	
	@property
	def arguments(self):
		return self.children[0].children
	
	def set_spacing(self, spacing):
		self.spacing = spacing
		return self
	
	def auto_parentheses(self):
		if len(self.children) == 0:
			return
		child = self.children[0] #sequence
		if len(child.children) == 0:
			return
		if self.parentheses_mode == "always":
			child.give_parentheses(True)
		elif self.parentheses_mode in ["weak", "strong"]:
			from ..expressions.operations import Operation, Add, Sub
			if len(child.children) > 1:
				child.give_parentheses(True)
			elif isinstance(child.children[0], (Add, Sub)):
				child.give_parentheses(True)
			else:
				if self.parentheses_mode == "strong":
					if isinstance(child.children[0], Operation):
						child.give_parentheses(True)
		elif self.parentheses_mode == "never":
			child.give_parentheses(False)
		else:
			raise ValueError(f"Unsupported parentheses mode {self.parentheses_mode}.")
		
	def compute(self, *args):
		if len(args) == 0:
			return self.rule(*map(lambda exp: exp.compute(), self.children[0].children))
		else:
			return self.rule(*args)
	
    #def __pow__(self, other):
    # Gotta do something about sin^2 etc   




from .expression_core import *


class Operation(Combiner):
	def __init__(self, symbol, symbol_glyph_length, *children, **kwargs):
		super().__init__(symbol, symbol_glyph_length, *children, **kwargs)

	def compute(self):
		result = self.children[0].compute()
		for child in self.children[1:]:
			result = self.eval_op(result, child.compute())
		return result


class Add(Operation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda x,y: x+y
		super().__init__("+", 1, *children, **kwargs)

	def auto_parentheses(self):
		for child in self.children:
			child.auto_parentheses()

	def is_negative(self):
		return self.children[0].is_negative()

class Sub(Operation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda x,y: x-y
		super().__init__("-", 1,*children, **kwargs)

	def auto_parentheses(self):
		self.children[0].auto_parentheses()
		for child in self.children[1:]:
			if isinstance(child, (Add, Sub)) or child.is_negative():
				child.give_parentheses()
			child.auto_parentheses()

	def is_negative(self):
		return self.children[0].is_negative()

class Mul(Operation):
	def __init__(self, *children, mode=None, **kwargs):
		self.eval_op = lambda x,y: x*y
		self.mode = algebra_config["multiplication_mode"] if mode is None else mode
		if self.mode == "dot":
			super().__init__("\\cdot", 1, *children, **kwargs)
		elif self.mode == "x":
			super().__init__("\\times", 1, *children, **kwargs)
		elif self.mode == "juxtapose":
			super().__init__("", 0, *children, **kwargs)
		else:
			raise ValueError(f"Invalid multiplication mode: {self.mode}. Mode must be dot, x, or juxtapose")

	def auto_parentheses(self): # should be more intelligent based on mode
		for child in self.children:
			if isinstance(child, (Add, Sub)) or child.is_negative():
				child.give_parentheses()
			child.auto_parentheses()

	def is_negative(self):
		return self.children[0].is_negative()

class Div(Operation):
	def __init__(self, *children, mode=None, **kwargs):
		self.eval_op = lambda x,y: x/y
		self.mode = algebra_config["division_mode"] if mode is None else mode
		if self.mode == "fraction":
			super().__init__("\\over", 1, *children, **kwargs)
		elif self.mode == "inline":
			super().__init__("\\div", 1, *children, **kwargs)
		else:
			raise ValueError(f"Invalid division mode: {self.mode}. Mode must be fraction or inline")

	def auto_parentheses(self):
		for child in self.children:
			if (isinstance(child, (Add, Sub, Mul, Div)) or child.is_negative()) and algebra_config["division_mode"] == "inline":
				child.give_parentheses()
			child.auto_parentheses()

	def is_negative(self):
		return self.children[0].is_negative() or self.children[1].is_negative()
	
	def compute(self):
		num = self.children[0].compute()
		den = self.children[1].compute()
		if den == 0:
			raise ZeroDivisionError
		if num % den == 0:
			return int(num / den)
		else:
			return float(num) / float(den)

class Pow(Operation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda x,y: x**y
		super().__init__("^", 0, *children, **kwargs)

	def auto_parentheses(self):
		assert len(self.children) == 2 #idc how to auto paren power towers
		if isinstance(self.children[0], Operation) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		for child in self.children:
			child.auto_parentheses()

	def is_negative(self):
		return False


class Negative(Expression):
	def __init__(self, child, **kwargs):
		self.children = [Smarten(child)]
		super().__init__(**kwargs)

	@tex
	def __str__(self):
		return "-" + str(self.children[0])

	def auto_parentheses(self):
		if isinstance(self.children[0], (Add, Sub)) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		self.children[0].auto_parentheses()

	def is_negative(self):
		return True

	def compute(self):
		return -self.children[0].compute()



from .timeline_core import *
from .timeline_variants import *
from ..actions.action_common import evaluate_

class Evaluate(AutoTimeline):
    def __init__(self, first_expression=None, mode="one at a time", number_mode="float", **kwargs):
        self.mode = mode
        self.number_mode = number_mode
        super().__init__(**kwargs)
        if first_expression is not None:
            self.add_expression_to_start(first_expression)

    def decide_next_action(self, index: int):
        last_exp = self.get_expression(index)
        leaves = last_exp.get_all_leaf_addresses()
        if leaves == ['']:
            return None
        leaves.sort(key=len, reverse=True)
        for leaf in leaves:
            try:
                twig = leaf[:-1]
                action = evaluate_(preaddress=twig)
                action.get_output_expression(last_exp)
                return action
            except ValueError:
                # This should mean that a subexpression cannot be computed due to the presence of a variable.
                # Perhaps we should make a custom Exception class for this so as not to accidentally catch others.
                pass
        return None



class Solve(AutoTimeline):
    def __init__(self, solve_for=Variable('x'), first_expression=None, preferred_side=None, auto_evaluate=True, **kwargs):
        super().__init__(**kwargs)
        if first_expression is not None:
            self.add_expression_to_start(first_expression)
        self.solve_for = solve_for
        self.auto_evaluate = auto_evaluate
        self.all_actions_to_try = []
        for maneuver_ in EquationManeuver.__subclasses__():
            self.all_actions_to_try += [
                maneuver_(),
                maneuver_().reverse(),
                maneuver_().flip(),
                maneuver_().reverse_flip()
            ]
    
    def decide_next_action(self, index:int):
        last_exp = self.get_expression(index)
        current_addresses = last_exp.get_addresses_of_subex(self.solve_for)
        assert len(current_addresses)==1, f"I don't know what to do if variable appears {len(current_addresses)} times"
        current_address = current_addresses[0]

        if self.auto_evaluate:
            try:
                result = (Evaluate(last_exp)).decide_next_action(0)
                if result is not None:
                    return result
            except:
                pass

        if current_address == '1':
            return swap_children_()
        if current_address == '0':
            return None

        if self.solve_for is not None:
            successful_outputs = []
            for maneuver in self.all_actions_to_try:
                try:
                    next_exp = maneuver.get_output_expression(last_exp)
                    new_address = next_exp.get_addresses_of_subex(self.solve_for)[0]
                    successful_outputs.append((maneuver, new_address))
                except:
                    pass
            
            if len(successful_outputs) == 0:
                return None
            shortest_result = min(successful_outputs, key=lambda p: len(p[1]))
            assert len(shortest_result[1]) <= len(current_address)
            return shortest_result[0]

        return None

    def set_solve_for(self, var):
        self.solve_for = var
        self.resume()




from .action_core import *
from ..expressions.expression_common import *
from MF_Tools.dual_compatibility import PI, FadeIn, FadeOut, Write, Uncreate
from .action_common import swap_children_


class AlgebraicAction(Action):
    def __init__(self, template1, template2, var_kwarg_dict={}, extra_addressmaps=[], **kwargs):
        super().__init__(**kwargs)
        self.template1 = template1
        self.template2 = template2
        self.var_kwarg_dict = var_kwarg_dict #{a:{"path_arc":PI}}
        self.extra_addressmaps = extra_addressmaps
        self.addressmap = None
    
    @preaddressfunc
    def get_output_expression(self, input_expression=None):
        var_dict = match_expressions(self.template1, input_expression)
        return self.template2.substitute(var_dict)
    
    @preaddressmap
    @autoparenmap
    def get_addressmap(self, input_expression=None):
        # Best overwritten in subclasses, but this gets the job done sometimes.
        # Actually, I think most subclasses will have a static addressmap, so I'll add this line at the start.
        if self.addressmap is not None:
            return self.addressmap

        addressmap = []
        def get_var_ad_dict(template):
            template_leaves = {
                template.get_subex(ad)
                for ad in template.get_all_leaf_addresses()
                }
            from ..expressions.variables import Variable
            variables = [var for var in template_leaves if isinstance(var, Variable)]
            return {var: template.get_addresses_of_subex(var) for var in variables}
        self.template1_address_dict = get_var_ad_dict(self.template1)
        self.template2_address_dict = get_var_ad_dict(self.template2)
        variables = self.template1_address_dict.keys() | self.template2_address_dict.keys()
        for var in variables:
            kwargs = self.var_kwarg_dict.get(var, {})
            if len(self.template1_address_dict[var]) == 1:
                addressmap += [[self.template1_address_dict[var][0], t2ad, kwargs] for t2ad in self.template2_address_dict[var]]
            elif len(self.template2_address_dict[var]) == 1:
                addressmap += [[t1ad, self.template2_address_dict[var][0], kwargs] for t1ad in self.template1_address_dict[var]]
            else:
                raise ValueError("I don't know what to do when a variable appears more than once on both sides. Please set addressmap manually.")
        addressmap += self.extra_addressmaps
        return addressmap

    def __repr__(self):
        return f"AlgebraicAction({self.template1}, {self.template2})"
    
    # def get_animation(self, *args, **kwargs):
    #     return super().get_animation(*args, auto_fade=True, auto_resolve_delay=0.1, **kwargs)

    def reverse(self):
        # swaps input and output templates
        self.template1, self.template2 = self.template2, self.template1


class EquationManeuver(AlgebraicAction):
    # Watch out, these cannot be preaddressed currently.
    # But I can't conceive of why you'd want to do that anyway.
    # Perhaps for a sequence of equations?
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def reverse(self):
        # swaps input and output templates
        self.template1, self.template2 = self.template2, self.template1
        # modifies addressmap accordingly
        # (swap order and negate path_arcs)
        for entry in self.addressmap:
            entry[0], entry[1] = entry[1], entry[0]
            if len(entry) == 3 and 'path_arc' in entry[2].keys():
                entry[2]['path_arc'] = -entry[2]['path_arc']
        return self

    def flip(self):
        # flips the equations of both templates
        s = swap_children_()
        self.template1 = s.get_output_expression(self.template1)
        self.template2 = s.get_output_expression(self.template2)
        # modifies addressmap accordingly
        # (swap first character of addresses and negate path_arcs)
        for entry in self.addressmap:
            if isinstance(entry[0], str):
                entry[0] = str(1-int(entry[0][0])) + entry[0][1:]
            if isinstance(entry[1], str):
                entry[1] = str(1-int(entry[1][0])) + entry[1][1:]
            if len(entry) == 3 and 'path_arc' in entry[2].keys():
                entry[2]['path_arc'] = -entry[2]['path_arc']
        return self

    def reverse_flip(self):
        return self.reverse().flip()
    
    






class alg_add_R(EquationManeuver):
    def __init__(self, **kwargs):
        super().__init__(
            a + b & c,
            a & c - b,
            **kwargs
        )
        self.addressmap = (
            ['01', '11', {'path_arc':PI}],
            ['0+', '1-', {'path_arc':PI}],
        )


class alg_add_L(EquationManeuver):
    def __init__(self, **kwargs):
        super().__init__(
            a + b & c,
            b & c - a,
            **kwargs
        )
        self.addressmap = (
            ['00', '11', {'path_arc':PI}],
            ['0+', '1-', {'path_arc':PI}]
        )
        # return (
        #     ['00', '11', {'path_arc':PI}],
        #     ['0+', FadeOut, {'run_time':0.5}],
        #     [FadeIn, '1-', {'run_time':0.5, 'delay':0.5}]
        # )


class alg_mul_R(EquationManeuver):
    def __init__(self, **kwargs):
        super().__init__(
            a * b & c,
            a & c / b,
            **kwargs
        )
        self.addressmap = (
            ['01', '11', {'path_arc':PI}],
            #['0*', FadeOut, {'run_time':0.5}],
            [Write, '1/', {'run_time':0.5, 'delay':0.5}]
        )
    
    def reverse(self):
        super().reverse()
        self.addressmap = (
            ['11', '01', {'path_arc':-PI}],
            ['1/', FadeOut, {'run_time':0.5, 'delay':0}]
        )
        return self


class alg_mul_L(EquationManeuver):
    def __init__(self, **kwargs):
        super().__init__(
            a * b & c,
            b & c / a,
            **kwargs
        )
        self.addressmap = (
            ['00', '11', {'path_arc':PI}],
            #['0*', FadeOut, {'run_time':0.5}],
            [Write, '1/', {'run_time':0.5, 'delay':0.5}]
        )

    def reverse(self):
        super().reverse()
        self.addressmap = (
            ['11', '00', {'path_arc':-PI}],
            ['1/', FadeOut, {'run_time':0.5, 'delay':0}]
        )
        return self


# class alg_pow_R(EquationManeuver):
#     def __init__(self, **kwargs):
#         super().__init__(
#             a**b & c,
#             a & rad(b)(c),
#             **kwargs
#         )








class difference_of_squares_(AlgebraicAction):
    def __init__(self, **kwargs):
        super().__init__(
            a**2 - b**2,
            (a+b)*(a-b),
            **kwargs
        )


from MF_Tools.dual_compatibility import (
	Text, dc_Tex as Tex,
	UP, DOWN, LEFT, RIGHT,
	GREEN, BLUE, ORANGE,
	Indicate,
	VGroup, VDict,
	Line
)
import numpy as np


def Smarten(input):
	from .expressions.expression_core import Expression
	from .actions.action_core import Action
	from .timelines.timeline_core import Timeline
	from .expressions.numbers import Integer, Real
	if isinstance(input, (Expression, Action, Timeline)):
		return input
	elif isinstance(input, int):
		return Integer(input)
	elif isinstance(input, float):
		return Real(input)
	else:
		raise NotImplementedError(f"Unsupported type {type(input)}")


def tex(func):
	def wrapper(expr, *args, **kwargs):
		pretex = func(expr, *args, **kwargs)
		if expr.parentheses:
			pretex = r"\left(" + pretex + r"\right)"
		return pretex
	return wrapper


def add_spaces_around_brackets(input_string): #GPT
	result = []
	i = 0
	length = len(input_string)

	while i < length:
		if input_string[i] == '{' or input_string[i] == '}':
			if i > 0 and input_string[i - 1] != ' ':
				result.append(' ')
			result.append(input_string[i])
			if i < length - 1 and input_string[i + 1] != ' ':
				result.append(' ')
		else:
			result.append(input_string[i])
		i += 1

	# Join the list into a single string and remove any extra spaces
	spaced_string = ''.join(result).split()
	return ' '.join(spaced_string)


def debug_smarttex(scene, smarttex, show_indices=True, show_addresses=True, show_submobjects=True):
	print("Debugging Expression:")
	print(smarttex)
	print("Length: ", len(smarttex))
	print("Type: ", type(smarttex))
	print("Number of children: ", len(smarttex.children))
	if show_indices:
		for index in range(len(smarttex)):
			index_text = Text(str(index), color=GREEN).next_to(smarttex, DOWN)
			scene.add(index_text)
			scene.play(Indicate(smarttex[0][index], color=GREEN))
			scene.remove(index_text)
	if show_addresses:
		for ad in smarttex.get_all_addresses():
			ad_text = Text(ad, color=ORANGE).next_to(smarttex, DOWN)
			subex_type = Text(type(smarttex.get_subex(ad)).__name__, color=ORANGE).next_to(ad_text, DOWN)
			scene.add(ad_text, subex_type)
			scene.play(Indicate(smarttex[ad], color=ORANGE))
			scene.remove(ad_text, subex_type)
	if show_addresses:
		for i, subm in enumerate(smarttex.submobjects[0]):
			subm_number = Text(str(i), color=BLUE).next_to(subm, DOWN)
			scene.add(subm_number)
			scene.play(Indicate(subm, color=BLUE))
			scene.remove(subm_number)


def match_expressions(template, expression):
	"""
		This function will either return a ValueError if the expression
		simply does not match the structure of the template, such as a missing
		operand or a plus in place of a times, or if they do match it will return
		a dictionary of who's who. For example,
		
		template:      (a*b)**n
		expression:    (4*x)**(3+y)
		return value:  {a:4, b:x, n:3+y}

		template:      n + x**5
		expression:    12 + x**3
		return value:  ValueError("Structures do not match at address 11, 5 vs 3")
		
		template:      x**n*x**m
		expression:    2**2*3**3
		return value:  ValueError("Conflicting matches for x: 2 and 3")

		Obviously this has to be recursive, but gee I am feeling a bit challenged atm...
		...
		Ok I think I've done it!
	"""
	from .expressions.variables import Variable
	# Leaf case
	if not template.children:
		if isinstance(template, Variable):
			return {template: expression}
		elif template.is_identical_to(expression):
			return {}
		else:
			raise ValueError("Expressions do not match")
	
	# Node case
	var_dict = {}
	if not isinstance(expression, type(template)):
		raise ValueError("Expressions do not match type")
	if not len(template.children) == len(expression.children):
		raise ValueError("Expressions do not match children length")
	for tc,ec in zip(template.children, expression.children):
		child_dict = match_expressions(tc,ec)
		matching_keys = child_dict.keys() & var_dict.keys()
		if any(not child_dict[key].is_identical_to(var_dict[key]) for key in matching_keys):
			raise ValueError("Conflicting matches for " + str(matching_keys))
		var_dict.update(child_dict)

	return var_dict


def random_number_expression(leaves=range(-5, 10), max_depth=3, max_children_per_node=2, **kwargs):
	import random
	from .expressions.numbers import Integer
	from .expressions.operations import Add, Sub, Mul, Div, Pow, Negative
	nodes = [Add, Sub, Mul, Pow]
	node = random.choice(nodes)
	def generate_child(current_depth):
		if np.random.random() < 1 / (current_depth + 1):
			return Integer(random.choice(leaves))
		else:
			return random_number_expression(leaves, max_depth - 1)
	def generate_children(current_depth, number_of_children):
		return [generate_child(current_depth) for _ in range(number_of_children)]
	if node == Add or node == Mul:
		children = generate_children(max_depth, random.choice(list(range(2,max_children_per_node+1))))
	elif node == Negative:
		children = generate_children(max_depth, 1)
	else:
		children = generate_children(max_depth, 2)
	return node(*children, **kwargs)


def create_graph(expr, node_size=0.5, horizontal_buff=1, vertical_buff=1.5, printing=False):
	def create_node(address):
		from .expressions.numbers import Integer, Real, Rational
		from .expressions.variables import Variable
		from .expressions.operations import Add, Sub, Mul, Div, Pow, Negative
		from .expressions.functions import Function
		from .expressions.sequences import Sequence
		from .expressions.relations import Equation, LessThan, LessThanOrEqualTo, GreaterThan, GreaterThanOrEqualTo
		type_to_symbol_dict = {
			Integer: lambda expr: str(expr.n),
			Real: lambda expr: expr.symbol if expr.symbol else str(expr),
			Rational: lambda expr: "\\div",
			Variable: lambda expr: expr.symbol,
			Add: lambda expr: "+",
			Sub: lambda expr: "-",
			Mul: lambda expr: "\\times",
			Div: lambda expr: "\\div",
			Pow: lambda expr: "\\hat{}",
			Negative: lambda expr: "-",
			Function: lambda expr: expr.symbol,
			Sequence: lambda expr: ",",
			Equation: lambda expr: "=",
			LessThan: lambda expr: "<",
			LessThanOrEqualTo: lambda expr: "\\leq",
			GreaterThan: lambda expr: ">",
			GreaterThanOrEqualTo: lambda expr: "\\geq",
		}
		subex = expr.get_subex(address)
		symbol = type_to_symbol_dict[type(subex)](subex)
		tex = Tex(symbol)
		# if tex.width > tex.height:
		# 	tex.scale_to_fit_width(node_size)
		# else:
		# 	tex.scale_to_fit_height(node_size)
		return tex
	addresses = expr.get_all_addresses()
	if printing: print(addresses)
	max_length = max(len(address) for address in addresses)
	layered_addresses = [
		[ad for ad in addresses if len(ad) == i]
		for i in range(max_length + 1)
	]
	if printing: print(layered_addresses)
	max_index = max(range(len(layered_addresses)), key=lambda i: len(layered_addresses[i]))
	max_layer = layered_addresses[max_index]
	max_width = len(max_layer)
	if printing: print(max_index, max_width, max_layer)
	Nodes = VDict({ad: create_node(ad) for ad in addresses})
	#Max_layer = VGroup(*[Nodes[ad] for ad in max_layer]).arrange(RIGHT,buff=horizontal_buff)
	def position_children(parent_address):
		parent = Nodes[parent_address]
		child_addresses = [ad for ad in layered_addresses[len(parent_address)+1] if ad[:-1] == parent_address]
		if printing: print(child_addresses)
		child_Nodes = VGroup(*[Nodes[ad] for ad in child_addresses]).arrange(RIGHT,buff=1)
		child_Nodes.move_to(parent.get_center()+DOWN*vertical_buff)
	for i in range(max_index, max_length):
		for ad in layered_addresses[i]:
			position_children(ad)
	def position_parent(child_address):
		sibling_Nodes = VGroup(*[Nodes[ad] for ad in layered_addresses[len(child_address)] if ad[:-1] == child_address[:-1]])
		parent_Node = Nodes[child_address[:-1]]
		parent_Node.move_to(sibling_Nodes.get_center()+UP*vertical_buff)
	for i in range(max_index, 0, -1):
		for ad in layered_addresses[i]:
			position_parent(ad)
	Edges = VGroup(*[
		Line(
			Nodes[ad[:-1]].get_bounding_box_point(DOWN),
			Nodes[ad].get_bounding_box_point(UP),
			buff=0.2, stroke_opacity=0.4
			)
		for ad in addresses if len(ad) > 0
		])
	return VGroup(Nodes, Edges)


"""
import MF_Algebra as MF
import sympy as sp
def expression_to_sympy(expr):
	expr = Smarten(expr)
	if isinstance(expr, MF.expressions.numbers.Number):
		if isinstance(expr, MF.expressions.numbers.Integer):
			return sp.Integer(expr.n)
		elif isinstance(expr, MF.expressions.numbers.Real):
			return sp.Float(expr.n)
		elif isinstance(expr, MF.expressions.numbers.Rational):
			return sp.Rational(expr.n, expr.d)
		else:
			raise NotImplementedError(f"Unsupported type {type(expr)}")
	elif isinstance(expr, MF.expressions.variables.Variable):
		return sp.Symbol(expr.symbol)
	elif isinstance(expr, MF.expressions.operations.Operation):
		if isinstance(expr, MF.expressions.operations.Add):
			return sp.Add(*map(expression_to_sympy, expr.children))
		elif isinstance(expr, MF.expressions.operations.Sub):
			return sp.Add(expression_to_sympy(expr.children[0]), -expression_to_sympy(expr.children[1]))
		elif isinstance(expr, MF.expressions.operations.Mul):
			return sp.Mul(*map(expression_to_sympy, expr.children))
		elif isinstance(expr, MF.expressions.operations.Div):
			return sp.Mul(expression_to_sympy(expr.children[0]), sp.Pow(expression_to_sympy(expr.children[1]), -1))
		elif isinstance(expr, MF.expressions.operations.Pow):
			return sp.Pow(expression_to_sympy(expr.children[0]), expression_to_sympy(expr.children[1]))
		elif isinstance(expr, MF.expressions.operations.Negative):
			return sp.Mul(-1, expression_to_sympy(expr.children[0]))
		else:
			raise NotImplementedError(f"Unsupported type {type(expr)}")
	elif isinstance(expr, MF.expressions.sequences.Sequence):
		raise NotImplementedError("Sequences not yet supported")
	elif isinstance(expr, MF.expressions.functions.Function):
		raise NotImplementedError("Functions not yet supported")
	elif isinstance(expr, MF.expressions.relations.Relation):
		raise NotImplementedError("Relations not yet supported")
	else:
		raise NotImplementedError(f"Unsupported type {type(expr)}")

"""


from .expression_core import *
from .numbers import Integer, Real, Rational
from .variables import Variable
from .operations import Add, Sub, Mul, Div, Pow, Negative
from .functions import Function
from .relations import Equation, LessThan, LessThanOrEqualTo, GreaterThan, GreaterThanOrEqualTo
import numpy as np


SmE = Expression
SmZ = Integer
SmQ = Rational
SmR = Real
SmAdd = Add
SmSub = Sub
SmMul = Mul
SmDiv = Div
SmPow = Pow
SmNeg = Negative
SmVar = Variable
SmFunc = Function
SmEq = Equation
SmLt = LessThan
SmLeq = LessThanOrEqualTo
SmGt = GreaterThan
SmGeq = GreaterThanOrEqualTo

a = SmVar("a")
b = SmVar("b")
c = SmVar("c")

n = SmVar("n")
m = SmVar("m")

r = SmVar("r")
t = SmVar("t")
u = SmVar("u")
v = SmVar("v")

x = SmVar("x")
y = SmVar("y")
z = SmVar("z")

alpha = SmVar("\\alpha")
beta = SmVar("\\beta")
gamma = SmVar("\\gamma")
theta = SmVar("\\theta")

e = SmR(np.e, "e")
pi = SmR(np.pi, "\\pi")
tau = SmR(np.pi*2, "\\tau")

f = SmFunc("f", 1)
g = SmFunc("g", 1)
h = SmFunc("h", 1)

sqrt = SmFunc("\\sqrt", 2, rule=lambda x: x**0.5, parentheses_mode="weak")

def rad(index):
    return SmFunc(f"\\sqrt[{index}]", 2+len(str(index)),
        rule=lambda x: x**(1/index),
        parentheses_mode="weak")

def log(base):
    return SmFunc(f"\\log_{base}", 3+len(str(base)),
    rule=lambda x: np.log(x)/np.log(base),
    parentheses_mode="weak")




from .expression_core import *


class Variable(Expression):
	def __init__(self, symbol, **kwargs):
		self.symbol = symbol
		self.children = []
		super().__init__(**kwargs)

	@tex
	def __str__(self):
		return self.symbol

	def is_identical_to(self, other):
		return type(self) == type(other) and self.symbol == other.symbol

	def compute(self):
		raise ValueError(f"Expression contains a variable {self.symbol}.")



from ...expressions.expression_core import *
from ...expressions.numbers import Real
from ...expressions.functions import Function
from ...expressions.relations import Equation
from ...utils import *


class Infinity(Real):
    def __init__(self, **kwargs):
        super().__init__(np.inf, "\\infty", **kwargs)



class Limit(Function):
    def __init__(self, variable, value, **kwargs):
        self.variable = Smarten(variable)
        self.value = Smarten(value)
        super().__init__(
            symbol = "\\lim_{" + str(self.variable) + "\\to" + str(self.value) + "}",
            symbol_glyph_length = 3 + len(self.variable) + 1 + len(self.value),
            parentheses_mode = "weak",
            **kwargs
        )
        self.children += [self.variable, self.value]


class Differential(Function):
    def __init__(self, **kwargs):
        super().__init__(
            symbol = "\\text{d} \\! \\!",
            symbol_glyph_length = 1,
            parentheses_mode="weak",
            **kwargs
        )


class Integral(Function):
    def __init__(self, lower_bound=None, upper_bound=None, **kwargs):
        self.bounds = [Smarten(bound) if bound is not None else None for bound in (lower_bound, upper_bound)]
        symbol = "\\int"
        symbol_glyph_length = 1
        for connector, bound in zip(["_", "^"], self.bounds):
            if bound is not None:
                symbol += connector + "{" + str(bound) + "}"
                symbol_glyph_length += len(bound)
        super().__init__(
            symbol,
            symbol_glyph_length,
            parentheses_mode="weak",
            **kwargs
        )


class Sum(Function):
    def __init__(self, variable, lower_bound, upper_bound, **kwargs):
        self.variable = Smarten(variable)
        self.lower_bound = Smarten(lower_bound)
        self.lower_equation = Equation(self.variable, self.lower_bound)
        self.upper_bound = Smarten(upper_bound)
        super().__init__(
            symbol = "\\sum_{" + str(self.lower_equation) + "}^{" + str(self.upper_bound) + "}",
            symbol_glyph_length = 1 + len(self.lower_equation) + len(self.upper_bound),
            parentheses_mode = "weak",
            **kwargs
        )
        self.children += [self.lower_equation, self.upper_bound]







from .action_core import *
from MF_Tools.dual_compatibility import AnimationGroup
from MF_Tools import TransformByGlyphMap


class AddressMapAction(Action):
    def __init__(self, *address_map, extra_animations=[], **kwargs):
        super().__init__(**kwargs)
        self.address_map = address_map
        self.extra_animations = extra_animations
    
    def get_animation(self, **kwargs):
        def animation(input_exp, output_exp=None):
            if output_exp is None:
                output_exp = self.get_output_expression(input_exp)
            return AnimationGroup(
                TransformByAddressMap(
                    input_exp,
                    output_exp,
                    *self.address_map,
                    **kwargs
                ),
                *self.extra_animations
            )
        return animation


class GlyphMapAction(Action):
    def __init__(self, *glyph_map, extra_animations=[], show_indices=False, **kwargs):
        super().__init__(**kwargs)
        self.glyph_map = glyph_map
        self.extra_animations = extra_animations
        self.show_indices = show_indices
    
    def get_animation(self, **kwargs):
        def animation(input_exp, output_exp=None):
            if output_exp is None:
                output_exp = self.get_output_expression(input_exp)
            return AnimationGroup(
                TransformByGlyphMap(
                    input_exp.mob,
                    output_exp.mob,
                    *self.glyph_map,
                    show_indices = self.show_indices,
                    **kwargs
                ),
                *self.extra_animations
            )
        return animation


class AnimationAction(Action):
    def __init__(self, animation, **kwargs):
        super().__init__(**kwargs)
        self.animation = animation # callable on two mobjects
    
    def get_animation(self, **kwargs):
        def animation(self, input_exp, output_exp=None):
            if output_exp is None:
                output_exp = self.get_output_expression(input_exp)
            return self.animation(input_exp.mob, output_exp.mob, **kwargs)
        return animation



from MF_Tools.transforms import TransformByGlyphMap



class TransformByAddressMap(TransformByGlyphMap):
    def __init__(self, expA, expB, *addressmap, **kwargs):
        self.addressmap = addressmap
        self.glyphmap = self.addressmap_to_glyphmap(expA, expB, addressmap)
        super().__init__(expA.mob, expB.mob, *self.glyphmap, **kwargs)
    
    def addressmap_to_glyphmap(self, expA, expB, addressmap):
        glyphmap = [
            (
                expA.get_glyphs(entry[0]) if isinstance(entry[0], str) else entry[0],
                expB.get_glyphs(entry[1]) if isinstance(entry[1], str) else entry[1],
                entry[2] if len(entry) > 2 else {}
            )
            for entry in addressmap
        ]
        return glyphmap







from .action_core import *


class ParallelAction(Action):
    def __init__(self, *actions, **kwargs):
        self.actions = actions
        super().__init__(**kwargs)
    
    def get_output_expression(self, input_expression=None):
        expr = input_expression
        for action in self.actions:
            expr = action.get_output_expression(expr)
        return expr
    
    def get_addressmap(self, input_expression=None):
        return sum([action.get_addressmap(input_expression) for action in self.actions], [])
    
    def __or__(self, other):
        if isinstance(other, ParallelAction):
            return ParallelAction(*self.actions, *other.actions)
        elif isinstance(other, Action):
            return ParallelAction(*self.actions, other)
        else:
            raise ValueError("Can only use | with other ParallelAction or Action")
    
    def __ror__(self, other):
        if isinstance(other, Action):
            return ParallelAction(other, *self.actions)
        else:
            return NotImplemented



class SequentialAction(Action):
    def __init__(self, *actions, **kwargs):
        self.actions = list(actions)
        super().__init__(**kwargs)
    
    def get_output_expression(self, input_expression=None):
        expr = input_expression
        for action in self.actions:
            expr = action.get_output_expression(expr)
        return expr

    def get_animations(self):
        return [action.get_animation() for action in self.actions]
    
    def get_animation(self, i):
        return self.actions[i].get_animation()
        
    def __rshift__(self, other):
        if isinstance(other, SequentialAction):
            return SequentialAction(*self.actions, *other.actions)
        elif isinstance(other, Action):
            return SequentialAction(*self.actions, other)
        else:
            return ValueError("Can only use >> with other SequentialAction or Action")
    
    def __rrshift__(self, other):
        if isinstance(other, SequentialAction):
            return SequentialAction(*other.actions, *self.actions)
        elif isinstance(other, Action):
            return SequentialAction(other, *self.actions)
        else:
            return ValueError("Can only use >> with other SequentialAction or Action")
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.actions[key]
        elif isinstance(key, slice):
            return SequentialAction(*self.actions[key])




from .expression_core import *
from .operations import Div
import numpy as np


class Number(Expression):
	def __init__(self, **kwargs):
		self.children = []
		super().__init__(**kwargs)

	def compute(self):
		return float(self)


class Integer(Number):
	def __init__(self, n, **kwargs):
		self.n = n
		super().__init__(**kwargs)

	@tex
	def __str__(self):
		return str(self.n)

	def __float__(self):
		return float(self.n)
	
	def compute(self):
		return self.n

	def is_identical_to(self, other):
		return type(self) == type(other) and self.n == other.n

	def is_negative(self):
		return self.n < 0

	@staticmethod
	def GCF(*smartnums):
		smartnums = list(map(Smarten, smartnums))
		nums = list(map(lambda N: N.n, smartnums))
		return Smarten(int(np.gcd.reduce(nums)))

	@staticmethod
	def LCM(*smartnums):
		smartnums = list(map(Smarten, smartnums))
		nums = list(map(lambda N: N.n, smartnums))
		return Smarten(int(np.lcm.reduce(nums)))

	def prime_factorization(self):
		...


class Real(Number):
	def __init__(self, x, symbol=None, **kwargs):
		self.x = x
		self.symbol = symbol
		super().__init__(**kwargs)

	@tex
	def __str__(self, decimal_places=4, use_decimal=False):
		if self.symbol and not use_decimal:
			return self.symbol
		rounded = round(self.x, decimal_places)
		if rounded == self.x:
			return str(rounded)
		else:
			return f"{self.x:.{decimal_places}f}" + r"\ldots"

	def __float__(self):
		return float(self.x)

	def is_identical_to(self, other):
		return type(self) == type(other) and self.x == other.x

	def is_negative(self):
		return self.x < 0


class Rational(Div):
	# Better to subclass Div than Number because 5/3 is no more a number than 5^3 or 5+3
	# Multiclassing is an option but seems to be more trouble than it's worth
	def __init__(self, a, b, **kwargs):
		if not isinstance(a, (Integer, int)):
			raise TypeError (f"Unsupported numerator type {type(a)}: {a}")
		if not isinstance(b, (Integer, int)):
			raise TypeError (f"Unsupported denominator type {type(b)}: {b}")
		super().__init__(a, b, **kwargs)

	def simplify(self):
		pass #idk will make later



from .calculus_core import *
from ...expressions.expression_common import *


inf = Infinity()
d = Differential()
dx = d(x)
dy = d(y)
dz = d(z)
dt = d(t)
dr = d(r)
theta = SmVar("\\theta")
dtheta = d(theta)

# None of this works currently due to a problem in Function lol
# a = Variable('a')
# b = Variable('b')
# SumRule = AlgebraicAction(d(a+b), d(a) + d(b))
# DifferenceRule = AlgebraicAction(d(a-b), d(a) - d(b))
# ProductRule = AlgebraicAction(d(a*b), d(a)*b + a*d(b))
# QuotientRule = AlgebraicAction(d(a/b), (d(a)*b - a*d(b))/b**2)
# PowerRule = AlgebraicAction(d(a**b), b*d(a**(b-1))*d(a))



# expressions.py
from MF_Tools.dual_compatibility import dc_Tex as Tex, MANIM_TYPE, VGroup
from ..utils import Smarten, tex, add_spaces_around_brackets
from copy import deepcopy


algebra_config = {
		"auto_parentheses": True,
		"multiplication_mode": "juxtapose",
		"division_mode": "fraction",
		"decimal_precision": 4,
		"always_color": {},
		"fast_paren_length": False
	}

class Expression:
	def __init__(self, parentheses=False, **kwargs):
		self.parentheses = parentheses
		if algebra_config["auto_parentheses"]:
			self.auto_parentheses()
		self._mob = None

	@property
	def mob(self):
		if self._mob is None:
			self.init_mob()
		return self._mob

	def init_mob(self, **kwargs):
		string = add_spaces_around_brackets(str(self))
		self._mob = Tex(string, **kwargs)
		self.set_color_by_subex(algebra_config["always_color"])
	
	def copy(self):
		return deepcopy(self)

	def __getitem__(self, key):
		if isinstance(key, str): # address of subexpressions, should return the glyphs corresponding to that subexpression
			if MANIM_TYPE == 'GL':
				return VGroup(*[self.mob[g] for g in self.get_glyphs(key)])
			elif MANIM_TYPE == 'CE':
				return VGroup(*[self.mob[0][g] for g in self.get_glyphs(key)])
			else:
				raise Exception(f"Unknown manim type: {MANIM_TYPE}")
		else: # preserve behavior of Tex indexing
			return self.mob.__getitem__(key)

	def get_all_addresses(self):
		# Returns the addresses of all subexpressions
		addresses = [""]
		for n in range(len(self.children)):
			for child_address in self.children[n].get_all_addresses():
				addresses.append(str(n)+child_address)
		return addresses
	
	def get_all_nonleaf_addresses(self):
		return sorted(list({a[:-1] for a in self.get_all_addresses() if a != ""}))
	
	def get_all_leaf_addresses(self):
		return sorted(list(set(self.get_all_addresses()) - set(self.get_all_nonleaf_addresses())))

	def get_subex(self, address_string):
		# Returns the Expression object corresponding to the subexpression at the given address.
		# Note that this is not a submobject of self! It is a different mobject probably not on screen,
		# it was just created to help create self.
		if address_string == "":
			return self
		elif int(address_string[0]) < len(self.children):
			return self.children[int(address_string[0])].get_subex(address_string[1:])
		else:
			raise IndexError(f"No subexpression of {self} at address {address_string} .")

	def is_identical_to(self, other):
		# Checks if they are equal as expressions. Implemented separately in leaves.
		return type(self) == type(other) and len(self.children) == len(other.children) \
			and all(self.children[i].is_identical_to(other.children[i]) for i in range(len(self.children)))

	def get_addresses_of_subex(self, subex):
		subex = Smarten(subex)
		addresses = []
		for ad in self.get_all_addresses():
			if self.get_subex(ad).is_identical_to(subex):
				addresses.append(ad)
		return addresses

	def get_glyphs_at_address(self, address):
		from .operations import Negative
		from .functions import Function
		start = 0
		for n,a in enumerate(address):
			parent = self.get_subex(address[:n])
			if parent.parentheses:
				start += parent.paren_length()
			if isinstance(parent, Combiner):
				for i in range(int(a)):
					sibling = parent.children[i]
					start += len(sibling)
					start += parent.symbol_glyph_length
			elif isinstance(parent, Negative):
				start += 1
			elif isinstance(parent, Function):
				start += parent.symbol_glyph_length
			else:
				raise ValueError(f"Invalid parent type: {type(parent)}. n={n}, a={a}.")
		end = start + len(self.get_subex(address))
		return list(range(start, end))

	def get_left_paren_glyphs(self, address):
		subex = self.get_subex(address)
		if not subex.parentheses:
			return []
		subex_glyphs = self.get_glyphs_at_address(address)
		start = subex_glyphs[0]
		end = start + subex.paren_length()
		return list(range(start, end))

	def get_right_paren_glyphs(self, address):
		subex = self.get_subex(address)
		if not subex.parentheses:
			return []
		subex_glyphs = self.get_glyphs_at_address(address)
		end = subex_glyphs[-1]
		start = end - subex.paren_length()
		return list(range(start+1, end+1))
	
	def get_exp_glyphs_without_parentheses(self, address):
		subex = self.get_subex(address)
		subex_glyphs = self.get_glyphs_at_address(address)
		if not subex.parentheses:
			return subex_glyphs
		else:
			paren_length = subex.paren_length()
			return subex_glyphs[paren_length:-paren_length]
	
	def get_op_glyphs(self, address):
		from .functions import Function
		subex = self.get_subex(address)
		if not isinstance(subex, (Combiner, Function)) or subex.symbol_glyph_length==0:
			return []
		subex_glyphs = self.get_glyphs_at_address(address)
		results = []
		turtle = subex_glyphs[0]
		if subex.parentheses:
			turtle += subex.paren_length()
		if isinstance(subex, Function):
			results += list(range(turtle, turtle + subex.symbol_glyph_length))
		elif isinstance(subex, Combiner):
			for child in subex.children[:-1]:
				turtle += len(child)
				results += list(range(turtle, turtle + subex.symbol_glyph_length))
				turtle += subex.symbol_glyph_length
		return results
	
	def get_glyphs(self, psuedoaddress):
		# Returns the list of glyph indices corresponding to the subexpression at the given address.
		# Can accept special characters at the end to trigger one of the special methods above.
		special_chars = "()_+-*/^,"
		found_special_chars = [(i,c) for (i,c) in enumerate(psuedoaddress) if c in special_chars]
		if len(found_special_chars) == 0:
			return self.get_glyphs_at_address(psuedoaddress)
		else:
			address = psuedoaddress[:found_special_chars[0][0]]
			results = []
			for i,c in found_special_chars:
				if c == "(":
					results += self.get_left_paren_glyphs(address)
				elif c == ")":
					results += self.get_right_paren_glyphs(address)
				elif c == "_":
					results += self.get_exp_glyphs_without_parentheses(address)
				elif c in "+-*/^,":
					results += self.get_op_glyphs(address)
			return sorted(set(results))

	def __len__(self):
		if MANIM_TYPE == 'GL':
			return len(self.mob)
		elif MANIM_TYPE == 'CE':
			return len(self.mob[0])
		else:
			raise Exception(f"Unknown manim type: {MANIM_TYPE}")

	def __neg__(self):
		from .operations import Negative
		return Negative(self)

	def __add__(self, other):
		from .operations import Add
		return Add(self, other)

	def __sub__(self, other):
		from .operations import Sub
		return Sub(self, other)

	def __mul__(self, other):
		from .operations import Mul
		return Mul(self, other)

	def __truediv__(self, other):
		from .operations import Div
		return Div(self, other)

	def __pow__(self, other):
		from .operations import Pow
		return Pow(self, other)

	def __radd__(self, other):
		from .operations import Add
		return Add(other, self)

	def __rsub__(self, other):
		from .operations import Sub
		return Sub(other, self)

	def __rmul__(self, other):
		from .operations import Mul
		return Mul(other, self)

	def __rtruediv__(self, other):
		from .operations import Div
		return Div(other, self)

	def __rpow__(self, other):
		from .operations import Pow
		return Pow(other, self)

	def __matmul__(self, expression_dict):
		return self.substitute(expression_dict)

	def __and__(self, other):
		from .relations import Equation
		return Equation(self, other)
	
	def __rand__(self, other):
		from .relations import Equation
		return Equation(other, self)

	def __rshift__(self, other):
		other = Smarten(other)
		from ..actions.action_core import Action
		from ..timelines.timeline_core import Timeline
		if isinstance(other, Expression):
			timeline = Timeline()
			timeline.add_expression_to_end(self).add_expression_to_end(other)
			return timeline
		elif isinstance(other, Action):
			timeline = Timeline()
			timeline.add_expression_to_end(self).add_action_to_end(other)
			return timeline
		else:
			return NotImplemented

	def __rrshift__(self, other):
		return Smarten(other).__rshift__(self)

	def is_negative(self):
		return False # catchall if not defined in subclasses

	def give_parentheses(self, parentheses=True):
		self.parentheses = parentheses
		self._mob = None # Don't init mob just yet, just mark it as needing to be reinitialized
		return self

	def clear_all_parentheses(self):
		for c in self.children:
			c.clear_all_parentheses()
		self.give_parentheses(False)
		return self

	def auto_parentheses(self):
		for child in self.children:
			child.auto_parentheses()
		return self
	
	def reset_parentheses(self):
		self.clear_all_parentheses()
		self.auto_parentheses()
		return self

	def paren_length(self):
		# Returns the number of glyphs taken up by the expression's potential parentheses.
		# Usually 1 but can be larger for larger parentheses.
		if algebra_config['fast_paren_length'] == True:
			return 1
		yes_paren = self.copy().give_parentheses(True)
		no_paren = self.copy().give_parentheses(False)
		num_paren_glyphs = len(yes_paren) - len(no_paren)
		assert num_paren_glyphs > 0 and num_paren_glyphs % 2 == 0
		return num_paren_glyphs // 2

	#Man these guys do not work correctly yet
	def nest(self, direction="right", recurse=True):
		if len(self.children) <= 2:
			return self
		else:
			if direction == "right":
				return type(self)(self.children[0], type(self)(*self.children[1:]).nest(direction, recurse))
			elif direction == "left":
				return type(self)(type(self)(*self.children[:-1]).nest(direction, recurse), self.children[-1])
			else:
				raise ValueError(f"Invalid direction: {direction}. Must be right or left.")

	def denest(self, denest_all = False, match_type = None):
		if len(self.children) <= 1:
			return self
		if match_type is None:
			match_type = type(self)
		new_children = []
		for child in self.children:
			if type(child) == match_type:
				for grandchild in child.children:
					new_children.append(grandchild.denest(denest_all, match_type))
			elif denest_all:
				new_children.append(child.denest(True, match_type))
			else:
				new_children.append(child)
		return type(self)(*new_children)

	def substitute_at_address(self, subex, address):
		from .functions import Function
		from .operations import Negative
		subex = Smarten(subex).copy() #?
		if len(address) == 0:
			return subex
		new_child = self.children[int(address[0])].substitute_at_address(subex, address[1:])
		new_children = self.children[:int(address[0])] + [new_child] + self.children[int(address[0])+1:]
		if isinstance(self, (Combiner, Negative)):
			return type(self)(*new_children)
		elif isinstance(self, Function):
			return Function(self.symbol, self.symbol_glyph_length, self.rule, self.algebra_rule, self.parentheses_mode)(*new_children)
		else:
			raise ValueError("Something went wrong here... and this whole method needs a rewrite / rethink lol")

	def substitute_at_addresses(self, subex, addresses):
		result = self.copy()
		for address in addresses:
			result = result.substitute_at_address(subex, address)
		return result

	def substitute(self, expression_dict):
		result = self.copy()
		dict_with_numbers = list(enumerate(expression_dict.items()))
		from .variables import Variable
		for i, (from_subex, to_subex) in dict_with_numbers:
			result = result.substitute_at_addresses(Variable(f"T_{i}"), result.get_addresses_of_subex(from_subex))
		for i, (from_subex, to_subex) in dict_with_numbers:
			result = result.substitute_at_addresses(to_subex, result.get_addresses_of_subex(Variable(f"T_{i}")))
		return result

	def set_color_by_subex(self, subex_color_dict):
		for subex, color in subex_color_dict.items():
			for ad in self.get_addresses_of_subex(subex):
				self[ad].set_color(color)
				if self.get_subex(ad).parentheses and not subex.parentheses:
					self[ad+"()"].set_color(self.color)
		return self

	def evaluate(self):
		return Smarten(self.compute())

	def __repr__(self):
		return type(self).__name__ + "(" + str(self) + ")"

	def get_all_subexpressions_of_type(self, expression_type):
		result = set()
		for child in self.children:
			if isinstance(child, expression_type):
				result |= set(child)
			else:
				result |= child.get_all_subexpressions_of_type(expression_type)
		return result
	
	def get_all_variables(self):
		from .variables import Variable
		return self.get_all_subexpressions_of_type(Variable)



class Combiner(Expression):
	def __init__(self, symbol, symbol_glyph_length, *children, **kwargs):
		self.symbol = symbol
		self.symbol_glyph_length = symbol_glyph_length
		self.children = list(map(Smarten,children))
		self.left_spacing = ""
		self.right_spacing = ""
		super().__init__(**kwargs)

	@tex
	def __str__(self, *args, **kwargs):
		joiner = self.left_spacing + self.symbol + self.right_spacing
		result = joiner.join(["{" + str(child) + "}" for child in self.children])
		return result

	def set_spacing(self, left_spacing, right_spacing):
		self.left_spacing = left_spacing
		self.right_spacing = right_spacing

	def get_op_glyphs(self, *args, **kwargs):
		return super().get_op_glyphs(*args, **kwargs)

	def get_glyphs_at_address(self, address):
		start = 0
		for n,a in enumerate(address):
			parent = self.get_subex(address[:n])
			if parent.parentheses:
				start += parent.paren_length()
			for i in range(int(a)):
				sibling = parent.children[i]
				start += len(sibling)
				start += parent.symbol_glyph_length
		end = start + len(self.get_subex(address))
		return list(range(start, end))
	
	def get_glyphs_at_addigit(self, addigit):
		child_index = int(addigit)
		start = 0
		if self.parentheses:
			start += self.paren_length()
		for sibling in self.children[:child_index]:
			start += len(sibling)
			start += self.symbol_glyph_length
		end = start + len(self.children[child_index])
		return list(range(start, end))



from .expression_core import *


class Relation(Combiner):
	def __init__(self, symbol, symbol_glyph_length, *children, **kwargs):
		super().__init__(symbol, symbol_glyph_length, *children, **kwargs)
	
	def compute(self):
		return all([self.eval_op(self.children[i], self.children[i+1]) for i in range(len(self.children)-1)])


class Equation(Relation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda X,Y: X.is_identical_to(Y)
		super().__init__("=", 1, *children, **kwargs)

class LessThan(Relation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda X,Y: X < Y
		super().__init__("<", 1, *children, **kwargs)

class GreaterThan(Relation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda X,Y: X > Y
		super().__init__(">", 1, *children, **kwargs)

class LessThanOrEqualTo(Relation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda X,Y: X <= Y
		super().__init__("\\leq", 1, *children, **kwargs)

class GreaterThanOrEqualTo(Relation):
	def __init__(self, *children, **kwargs):
		self.eval_op = lambda X,Y: X >= Y
		super().__init__("\\geq", 1, *children, **kwargs)




from .timeline_core import *


class AutoTimeline(Timeline):
    def __init__(self, auto_fill=True, **kwargs):
        super().__init__(**kwargs)
        self.auto_fill = auto_fill
    
    def set_expression(self, index: int, expression: Expression):
        super().set_expression(index, expression)
        if self.auto_fill and self.get_action(index) is None:
            next_action = self.decide_next_action(index)
            if next_action is not None:
                self.set_action(index, next_action)
        return self
    
    def set_action(self, index: int, action: Action):
        super().set_action(index, action)
        if self.auto_fill and self.get_expression(index+1) is None:
            next_expression = action.get_output_expression(self.get_expression(index))
            self.set_expression(index+1, next_expression)
        return self
    
    def decide_next_action(self, index: int) -> Action:
        # Implement in subclasses. Return None if finished.
        return None

    def resume(self):
        self.add_action_to_end(self.decide_next_action(self.current_exp_index))






